/*
 * ====================================================================
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 */
package org.apache.hc.core5.http2.frame;

/**
 * Standard HTTP/2 frame flags.
 *
 * @since 5.0
 */
public enum FrameFlag {

    END_STREAM    (0x01),
    ACK           (0x01),
    END_HEADERS   (0x04),
    PADDED        (0x08),
    PRIORITY      (0x20);

    final int value;

    FrameFlag(final int value) {
        this.value = value;
    }

    public int getValue() {
        return value;
    }

    public static int of(final FrameFlag... flags) {
        int value = 0;
        for (final FrameFlag flag: flags) {
            value |= flag.value;
        }
        return value;
    }

}
